//#############################################################################
// Bob's Action Racing for Arduboy
// (c) 2016 by hackefuffel
// http://hackefuffel.com
// License: CC BY-NC-SA 4.0
// Credits: Bob Fungus, @chris_muetze, rou
//#############################################################################

#include <Arduboy.h>

// Images
#include "images.h"

//#############################################################################
// DEFINES
//#############################################################################

#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define NUM_LINES 5
#define NUM_CELLS 13
#define NUM_CELLS_LEFTRIGHT 5
#define CELL_SIZE 16
#define MIDDLE (128 / 2)
#define POS_Y (64 - 20)
#define SPEED_STEP (5 * 60)
#define MAX_SPEED 3
#define SCORE_GUY 60

#define RAND_FENCE 25
#define RAND_ROCK 8
#define RAND_TREE 10
#define RAND_SPIKE 20
#define RAND_GUY 6

#define PLAYER_STEER_SPEED 1
#define PLAYER_MIN_X 0
#define PLAYER_MAX_X (128 - CELL_SIZE)

#define NUM_HIGHSCORE 3
#define HIGHSCORE_BASE_ADDR 0

//#############################################################################
// TYPES
//#############################################################################

typedef unsigned char u8;
typedef signed char i8;
typedef unsigned short u16;
typedef signed short i16;
//typedef unsigned int u32;
typedef int i32;

//=============================================================================
// Game State
//=============================================================================
enum EState
{
    E_LOGO,
    E_GAME,
    E_DEAD,
    E_HIGHSCORE
};

//=============================================================================
// Cell Type
//=============================================================================
enum ECellType
{
    E_EMPTY,
    E_STREET_LEFT,
    E_STREET_MIDDLE,
    E_STREET_RIGHT,
    E_ROCK,
    E_TREE,
    E_FENCE,
    E_SPIKE,
    E_GUY,
    E_BODY
};

//=============================================================================
// Car state
//=============================================================================
enum ECarState
{
    E_NORMAL,
    E_LEFT,
    E_RIGHT,
    E_CRASH
};

//=============================================================================
// Button states
//=============================================================================
enum EButtonEvents
{
    E_NONE,
    E_PRESSED,
    E_RELEASED
};

//=============================================================================
// Player
// all infos about the player/car
//=============================================================================
struct TPlayer
{
    u8 pos;
    i8 car;
};

//=============================================================================
// Line
// a level line
//=============================================================================
struct TLine
{
    u16 cells[NUM_CELLS];
    u8 offset;
    i8 scroll;
};

//=============================================================================
// Level
//=============================================================================
struct TLevel
{
    i8 targetStepsX;
    u8 targetStepsY;
    u8 currentX;
    u8 currentType;
    u32 score;
    u32 tick;
    u32 nextspeed;
    u8 speed;
    u8 wall;
    u8 spike;

    TLine lines[NUM_LINES];
};

//=============================================================================
// Buttons
//=============================================================================
struct TButtons
{
    u8 lastA;
    u8 lastB;
    u8 eventA;
    u8 eventB;
};

//=============================================================================
// Highscore
//=============================================================================
struct THighscore
{
    u32 scores[NUM_HIGHSCORE];
    i8 last;
};

//=============================================================================
// 
//=============================================================================
struct TSound
{
    u32 freq;
    u32 duration;
    u32 ticks;
};


//#############################################################################
// GLOBALS
//#############################################################################

EState State = E_LOGO;
TPlayer Player;
TLevel Level;
TButtons Buttons;
THighscore Highscore;
TSound Sound;

const unsigned char* SpritesCar[] = 
{
    car_normal,
    car_left,
    car_right,
    car_crash
};

const unsigned char* SpritesStreet[] =
{
    street_left,
    street_normal,
    street_right
};

//#############################################################################
// IMPLEMENTATION
//#############################################################################

//=============================================================================
// load highscore
//=============================================================================
void bobLoadHighscore()
{
    int address = HIGHSCORE_BASE_ADDR;
    uint8_t m0 = EEPROM.read(address++);
    uint8_t m1 = EEPROM.read(address++);
    uint8_t m2 = EEPROM.read(address++);

    if ((m0 == 'b') && (m1 == 'o') && (m2 == 'b'))
    {
        for (int i = 0; i < NUM_HIGHSCORE; ++i)
        {
            Highscore.scores[i] = EEPROM.read(address++) | (EEPROM.read(address++) << 8)
                | (EEPROM.read(address++) << 16) | (EEPROM.read(address++) << 24);
        }
    }
    else
    {
        for (int i = 0; i < NUM_HIGHSCORE; ++i)
        {
            Highscore.scores[i] = 0;
        }
    }
}

//=============================================================================
// save highscore
//=============================================================================
void bobSaveHighscore()
{
    int address = HIGHSCORE_BASE_ADDR;
    EEPROM.write(address++, 'b');
    EEPROM.write(address++, 'o');
    EEPROM.write(address++, 'b');
    
    for (int i = 0; i < NUM_HIGHSCORE; ++i)
    {
        EEPROM.write(address++, (Highscore.scores[i] & 0x000000FF) >> 0);
        EEPROM.write(address++, (Highscore.scores[i] & 0x0000FF00) >> 8);
        EEPROM.write(address++, (Highscore.scores[i] & 0x00FF0000) >> 16);
        EEPROM.write(address++, (Highscore.scores[i] & 0xFF000000) >> 24);
    }
}

//=============================================================================
// init the game
//=============================================================================
void bobInit()
{
    randomSeed(Level.currentX);
    Buttons.lastA = E_RELEASED;
    Buttons.lastB = E_RELEASED;
    Buttons.eventA = E_RELEASED;
    Buttons.eventB = E_RELEASED;
    
    Sound.ticks = 0;

    bobLoadHighscore();   
}

//=============================================================================
// play sound
//=============================================================================
void bobPlaySound(Arduboy& arduboy, u32 freq, u32 dur, u32 ticks, bool override)
{
    if (override || !Sound.ticks)
    {
        Sound.duration = dur;
        Sound.freq = freq;
        Sound.ticks = ticks;
    }
}

//=============================================================================
// play sound
//=============================================================================
void bobTickSound(Arduboy& arduboy)
{
    if (Sound.ticks)
    {
        arduboy.tunes.tone(Sound.freq, Sound.duration);
        --Sound.ticks;
    }
}

//=============================================================================
// enter highscore
//=============================================================================
void bobHighscore(u32 score)
{
    i8 rank = -1;
    
    if (score > Highscore.scores[NUM_HIGHSCORE - 1])
    {
        for (int i = 0; i < NUM_HIGHSCORE; ++i)
        {
            if (score > Highscore.scores[i])
            {
                rank = i;
                break;
            }
        }
    }

    if (rank >= 0)
    {
        int s;
        for (int i = rank; i < NUM_HIGHSCORE; ++i)
        {
            s = Highscore.scores[i];
            Highscore.scores[i] = score;
            score = s;
        }
    }
    Highscore.last = rank;

    if (rank >= 0)
    {
        bobSaveHighscore();
    }
}

//=============================================================================
// init a line
//=============================================================================
void bobInitLine(u8 index, i8 scroll)
{
    // the initial scroll is given (usually jump back)
    Level.lines[index].scroll = scroll;

    // init empty and street. street is always in the same
    // cells. middle street is drawn every n times only.
    int cur = 0;
    for (int i = 0; i < NUM_CELLS_LEFTRIGHT; ++i)
    {
        Level.lines[index].cells[cur++] = E_EMPTY;
    }
    
    Level.lines[index].cells[cur++] = E_STREET_LEFT;
    Level.lines[index].cells[cur++] = Level.currentType++ == 0 ? E_STREET_MIDDLE : E_EMPTY;
    Level.lines[index].cells[cur++] = E_STREET_RIGHT;
    
    if (Level.currentType == 2) Level.currentType = 0;
    
    for (int i = 0; i < NUM_CELLS_LEFTRIGHT; ++i)
    {
        Level.lines[index].cells[cur++] = E_EMPTY;
    }

    bool skip = false;

    // make it a fence line. nothing else in that line
    if (Level.wall > 0)
    {
        --Level.wall;
    }

    if((Level.wall == 0) && (random(0, RAND_FENCE) == 0))
    {
        for (int i = 0; i < NUM_CELLS_LEFTRIGHT - 1; ++i)
        {
            Level.lines[index].cells[i] = E_FENCE;
        }

        for (int i = 0; i < NUM_CELLS_LEFTRIGHT - 1; ++i)
        {
            Level.lines[index].cells[NUM_CELLS - 1 - i] = E_FENCE;
        }

        Level.wall = 5;

        skip = true;
    }

    // spike
    if (!skip)
    {
        if (Level.spike > 0)
        {
            --Level.spike;
        }

        if(( Level.spike == 0 ) && (random(0, RAND_SPIKE) == 0) )
        {
            Level.lines[index].cells[NUM_CELLS_LEFTRIGHT + 1] = E_SPIKE;
            skip = true;
            Level.spike = 5;
        }
    }

    // rock
    if (!skip)
    {
        if (random(0, RAND_ROCK) == 0)
        {
            int idx = random(0, NUM_CELLS_LEFTRIGHT);
            if (Level.lines[index].cells[idx] == E_EMPTY)
            {
                Level.lines[index].cells[idx] = E_ROCK;
            }
        }
        else if (random(0, RAND_ROCK) == 0)
        {
            int idx = NUM_CELLS_LEFTRIGHT + 3 + random(0, NUM_CELLS_LEFTRIGHT - 1);
            if (Level.lines[index].cells[idx] == E_EMPTY)
            {
                Level.lines[index].cells[idx] = E_ROCK;
            }
        }
    }

    // tree
    if (!skip)
    {
        if (random(0, RAND_TREE) == 0)
        {
            int idx = random(0, NUM_CELLS_LEFTRIGHT);
            if (Level.lines[index].cells[idx] == E_EMPTY)
            {
                Level.lines[index].cells[idx] = E_TREE;
            }
        }
        else if (random(0, RAND_TREE) == 0)
        {
            int idx = NUM_CELLS_LEFTRIGHT + 3 + random(0, NUM_CELLS_LEFTRIGHT - 1);
            if (Level.lines[index].cells[idx] == E_EMPTY)
            {
                Level.lines[index].cells[idx] = E_TREE;
            }
        }
    }

    // guy
    if (!skip)
    {
        if (random(0, RAND_GUY) == 0)
        {
            bool found = false;
            do
            {
                int idx = random(0, NUM_CELLS);
                if ((Level.lines[index].cells[idx] == E_EMPTY) || (Level.lines[index].cells[idx] == E_STREET_MIDDLE))
                {
                    Level.lines[index].cells[idx] = E_GUY;
                    found = true;
                }
            } while (!found);
        }
    }


    // offset the line depending on current movement. calc
    // a new movement here if the old is done.
    
    if (Level.targetStepsY == 0)
    {
        Level.targetStepsX = random(-2, 3);
        Level.targetStepsY = random(4, 10);
    }

    i16 newX = Level.currentX + Level.targetStepsX;
    if (newX < 0)
    {
        newX = 0;
    }
    else if (newX > CELL_SIZE * 5)
    {
        newX = CELL_SIZE * 5;
    }

    Level.currentX = newX;
    Level.lines[index].offset = Level.currentX;
    --Level.targetStepsY;
}

//=============================================================================
// new game
//=============================================================================
void bobInitGame()
{
    Level.tick = 0;

    Player.pos = 56;
    Player.car = E_NORMAL;

    Level.targetStepsX = 0;
    Level.targetStepsY = 3;
    Level.currentX = 40;
    Level.currentType = 0;
    Level.score = 0;
    Level.nextspeed = 0;
    Level.speed = 0;
    Level.tick = 0;
    Level.spike = 8;
    Level.wall = 6;

    u8 start = SCREEN_HEIGHT - CELL_SIZE * NUM_LINES;
    for (int i = 0; i < NUM_LINES; ++i)
    {
        bobInitLine(i, start);
        start += CELL_SIZE;
    }
}

//=============================================================================
// switch state
//=============================================================================
void bobState(Arduboy& arduboy, EState newState)
{
    switch (newState)
    {
    case E_LOGO:
        bobPlaySound(arduboy, 600, 200, 5, true);
        break;

    case E_GAME:
        bobPlaySound(arduboy, 600, 200, 15, true);
        bobInitGame();
        break;

    case E_HIGHSCORE:
        bobPlaySound(arduboy, 600, 200, 5, true);
        break;

    case E_DEAD:
        bobPlaySound(arduboy, 500, 2, 25, true);
        Player.car = E_CRASH;
        bobHighscore(Level.score);
        break;
    }

    State = newState;
}

//=============================================================================
// logo
//=============================================================================
void bobLogo(Arduboy& arduboy)
{
    arduboy.clear();
    arduboy.drawBitmap(0, 24, logo, 128, 16, 1);

    if (Buttons.eventB == E_PRESSED)
    {
        bobState(arduboy, E_GAME);
    }
}

//=============================================================================
// over
//=============================================================================
void bobOver(Arduboy& arduboy)
{
    arduboy.clear();
    arduboy.drawBitmap(0, 2, logo, 128, 16, 1);

    for (int i = 0; i < NUM_HIGHSCORE; ++i)
    {
        if ( (i != Highscore.last) || ( Level.tick % 40 > 20 ) )
        {
            arduboy.setCursor(30, 24 + i * 12);
            arduboy.print(i + 1);
            arduboy.print('.');

            arduboy.setCursor(60, 24 + i * 12);
            arduboy.print(Highscore.scores[i]);
        }
    }

    if (Buttons.eventB == E_PRESSED)
    {
        bobState(arduboy, E_GAME);
    }
}

//=============================================================================
// move forward
//=============================================================================
u8 bobMoveForward()
{
    switch (Level.speed)
    {
    case 0:
        return Level.tick % 2;

    case 1:
        return 1;

    case 2:
        return (Level.tick % 2) ? 1 : 2;

    case 3:
        return 2;

    case 4:
        return (Level.tick % 2) ? 2 : 3;

    default:
    case 5:
        return 3;
    }
}

//=============================================================================
// collision stuff
//=============================================================================
bool bobCollide(Arduboy& arduboy)
{
    // each line
    for (int line = 0; line < NUM_LINES; ++line)
    {
        // test y (avoids checking all those cells)
        int y = Level.lines[line].scroll;
        
        if( ( POS_Y < y + CELL_SIZE ) && ( POS_Y > y - (CELL_SIZE / 2) ) )
        {
            // check those cells
            for (int cell = 0; cell < NUM_CELLS; ++cell)
            {
                int x = Level.lines[line].offset - CELL_SIZE * NUM_CELLS_LEFTRIGHT + cell * CELL_SIZE;

                int dx = x > Player.pos ? x - Player.pos : Player.pos - x;
                if (dx < CELL_SIZE)
                {
                    switch (Level.lines[line].cells[cell])
                    {
                    case E_STREET_LEFT:
                    case E_STREET_MIDDLE:
                    case E_STREET_RIGHT:
                    case E_EMPTY:
                    case E_BODY:
                        break;

                    case E_GUY:
                        Level.lines[line].cells[cell] = E_BODY;
                        Level.nextspeed = 0;
                        if (Level.speed) --Level.speed;
                        Level.score += SCORE_GUY;
                        bobPlaySound(arduboy, 500, 2, 15, true);
                        return false;
                        
                    default:
                        return true;
                    }
                }
            }
        }
    }

    return false;
}

//=============================================================================
// move all elements and perform collision detection
//=============================================================================
void bobMove(Arduboy& arduboy)
{
    if (arduboy.pressed(LEFT_BUTTON))
    {
        Player.car = E_LEFT;
        
        int newX = Player.pos - PLAYER_STEER_SPEED;
        if (newX < PLAYER_MIN_X)
        {
            newX = PLAYER_MIN_X;
        }
        Player.pos = (u8)newX;
    }
    else if (arduboy.pressed(RIGHT_BUTTON))
    {
        Player.car = E_RIGHT;

        int newX = Player.pos + PLAYER_STEER_SPEED;
        if (newX > PLAYER_MAX_X)
        {
            newX = PLAYER_MAX_X;
        }
        Player.pos = (u8)newX;
    }
    else
    {
        Player.car = E_NORMAL;
    }

    u8 step = bobMoveForward();

    for (int i = 0; i < NUM_LINES; ++i)
    {
        Level.lines[i].scroll += step;
        if (Level.lines[i].scroll >= SCREEN_HEIGHT)
        {
            bobInitLine(i, Level.lines[i].scroll - NUM_LINES * CELL_SIZE);
        }
    }

    Level.score += step;
    Level.nextspeed++;

    if (Level.nextspeed > SPEED_STEP)
    {
        Level.nextspeed = 0;
        ++Level.speed;
        if (Level.speed > MAX_SPEED)
        {
            Level.speed = MAX_SPEED;
        }
    }

    bobPlaySound(arduboy, 420 + Level.speed * 40, 9, 1, false);

    if (bobCollide(arduboy))
    {
        bobState(arduboy, E_DEAD);
    }
}

//=============================================================================
// 
//=============================================================================
void bobDrawLine(Arduboy& arduboy, int index)
{
    int x = Level.lines[index].offset - CELL_SIZE * NUM_CELLS_LEFTRIGHT;
    for (int i = 0; i < NUM_CELLS; ++i)
    {
        // clip
        if ((x > -CELL_SIZE) && (x < SCREEN_WIDTH))
        {
            switch (Level.lines[index].cells[i])
            {
            case E_STREET_LEFT:
            case E_STREET_MIDDLE:
            case E_STREET_RIGHT:
                arduboy.drawBitmap(x, Level.lines[index].scroll, SpritesStreet[Level.lines[index].cells[i] - E_STREET_LEFT], CELL_SIZE, CELL_SIZE, 1);
                break;

            case E_FENCE:
                arduboy.drawBitmap(x, Level.lines[index].scroll, wall, CELL_SIZE, CELL_SIZE, 1);
                break;

            case E_TREE:
                arduboy.drawBitmap(x, Level.lines[index].scroll, tree, CELL_SIZE, CELL_SIZE, 1);
                break;

            case E_ROCK:
                arduboy.drawBitmap(x, Level.lines[index].scroll, rock, CELL_SIZE, CELL_SIZE, 1);
                break;

            case E_SPIKE:
                arduboy.drawBitmap(x, Level.lines[index].scroll, oil, CELL_SIZE, CELL_SIZE, 1);
                break;

            case E_GUY:
            {
                int t = Level.tick % 80;
                if (t >= 60)
                {
                    arduboy.drawBitmap(x, Level.lines[index].scroll, guy_001, CELL_SIZE, CELL_SIZE, 1);
                }
                else if (t >= 40)
                {
                    arduboy.drawBitmap(x, Level.lines[index].scroll, guy_002, CELL_SIZE, CELL_SIZE, 1);
                }
                else if (t >= 20)
                {
                    arduboy.drawBitmap(x, Level.lines[index].scroll, guy_003, CELL_SIZE, CELL_SIZE, 1);
                }
                else
                {
                    arduboy.drawBitmap(x, Level.lines[index].scroll, guy_002, CELL_SIZE, CELL_SIZE, 1);
                }

                break;
            }

            case E_BODY:
                arduboy.drawBitmap(x, Level.lines[index].scroll, body, CELL_SIZE, CELL_SIZE, 1);
                break;
            }
        }

        x += CELL_SIZE;
    }
}

//=============================================================================
// 
//=============================================================================
void bobDrawGame(Arduboy& arduboy)
{
    for (int i = 0; i < NUM_LINES; ++i)
    {
        bobDrawLine(arduboy, i);
    }

    // draw car
    arduboy.drawBitmap(Player.pos, POS_Y, SpritesCar[Player.car], 16, 16, 1);
}

//=============================================================================
// game
//=============================================================================
void bobGame(Arduboy& arduboy)
{
    bobMove(arduboy);
    arduboy.clear();
    bobDrawGame(arduboy);
}

//=============================================================================
// dead
//=============================================================================
void bobDead(Arduboy& arduboy)
{
    arduboy.clear();
    bobDrawGame(arduboy);

    arduboy.drawFastHLine(0, 19, 128, 1);
    arduboy.drawFastHLine(0, 36, 128, 1);
    arduboy.drawBitmap(Player.pos, POS_Y, SpritesCar[E_CRASH], 16, 16, 1);
    arduboy.fillRect(0, 20, 128, 16, 0);
    arduboy.drawBitmap(22, 20, game_over, 84, 16, 1);

    if (Buttons.eventB == E_PRESSED)
    {
        bobState(arduboy, E_HIGHSCORE);
    }
}

//=============================================================================
// 
//=============================================================================
void bobButtons(Arduboy& arduboy)
{
    // button down
    if (arduboy.pressed(A_BUTTON) && (Buttons.lastA == 0))
    {
        Buttons.eventA = E_PRESSED;
    }
    Buttons.lastA = arduboy.pressed(A_BUTTON);

    // button down
    if (arduboy.pressed(B_BUTTON) && (Buttons.lastB == 0))
    {
        Buttons.eventB = E_PRESSED;
    }
    Buttons.lastB = arduboy.pressed(B_BUTTON);
    
}

//=============================================================================
// 
//=============================================================================
void bobGameLoop(Arduboy& arduboy)
{
    bobButtons(arduboy);

    switch (State)
    {
    case E_LOGO:
        bobLogo(arduboy);
        break;

    case E_GAME:
        bobGame(arduboy);
        break;

    case E_DEAD:
        bobDead(arduboy);
        break;

    case E_HIGHSCORE:
        bobOver(arduboy);
        break;
    }

    ++Level.tick;
    bobTickSound(arduboy);

    Buttons.eventA = E_NONE;
    Buttons.eventB = E_NONE;
}
